(* $Id: varcc.ml4,v 1.2 2003/06/02 03:42:43 eric Exp $ *)

(* Compile a list of variant tags into CPP defines *) 

open StdLabels

(* hash_variant, from ctype.ml *)

let hash_variant s =
  let accu = ref 0 in
  for i = 0 to String.length s - 1 do
    accu := 223 * !accu + Char.code s.[i]
  done;
  accu := !accu land (1 lsl 31 - 1);
  if !accu > 0x3FFFFFFF then !accu - 1 lsl 31 else !accu

open Genlex

let lexer = make_lexer ["type"; "public"; "private"; "="; "["; "]"; "`"; "|"]

let may_string (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (String s) -> Stream.junk strm__; s
  | _ -> ""

let may_bar (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "|") -> Stream.junk strm__; ()
  | _ -> ()

let rec ident_list (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "`") ->
      Stream.junk strm__;
      begin match Stream.peek strm__ with
        Some (Ident x) ->
          Stream.junk strm__;
          let trans =
            try may_string strm__ with
              Stream.Failure -> raise (Stream.Error "")
          in
          let _ =
            try may_bar strm__ with
              Stream.Failure -> raise (Stream.Error "")
          in
          (x, trans) :: ident_list strm__
      | _ -> raise (Stream.Error "")
      end
  | _ -> []

let static = ref false
let may_public (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "public") -> Stream.junk strm__; true
  | Some (Kwd "private") -> Stream.junk strm__; false
  | _ -> not !static

open Printf

let hashes = Hashtbl.create 57

let declaration ~hc ~cc (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "type") ->
      Stream.junk strm__;
      let public =
        try may_public strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some (Ident name) ->
          Stream.junk strm__;
          begin match Stream.peek strm__ with
            Some (Kwd "=") ->
              Stream.junk strm__;
              let prefix =
                try may_string strm__ with
                  Stream.Failure -> raise (Stream.Error "")
              in
              begin match Stream.peek strm__ with
                Some (Kwd "[") ->
                  Stream.junk strm__;
                  let _ =
                    try may_bar strm__ with
                      Stream.Failure -> raise (Stream.Error "")
                  in
                  let tags =
                    try ident_list strm__ with
                      Stream.Failure -> raise (Stream.Error "")
                  in
                  begin match Stream.peek strm__ with
                    Some (Kwd "]") ->
                      Stream.junk strm__;
                      let suffix =
                        try may_string strm__ with
                          Stream.Failure -> raise (Stream.Error "")
                      in
                      let oh x = fprintf hc x
                      and oc x = fprintf cc x in
                      let first = ref true in
                      List.iter tags
                        ~f:(
                           fun (tag, _) ->
                             let hash = hash_variant tag in
                             try
                               let tag' = Hashtbl.find hashes hash in
                               if tag <> tag' then
                                 failwith
                                   (String.concat ~sep:" "
                                      ["Doublon tag:"; tag; "and"; tag'])
                             with
                               Not_found ->
                                 Hashtbl.add hashes hash tag;
                                 if !first then
                                   begin
                                     oh "/* %s : tags and macros */\n" name;
                                     first := false
                                   end;
                                 oh "#define MLTAG_%s\tVal_int(%d)\n" tag
                                    hash);
                      let ctag tag trans =
                        if trans <> "" then trans
                        else
                          let tag =
                            if tag.[0] = '_' then
                              String.sub tag ~pos:1
                                ~len:(String.length tag - 1)
                            else tag
                          in
                          match
                            if prefix = "" then None, ""
                            else
                              Some prefix.[String.length prefix - 1],
                              String.sub prefix ~pos:0
                                ~len:(String.length prefix - 1)
                          with
                            Some '#', prefix ->
                              prefix ^ String.uncapitalize tag ^ suffix
                          | Some '^', prefix ->
                              prefix ^ String.uppercase tag ^ suffix
                          | _ -> prefix ^ tag ^ suffix
                      and cname = String.capitalize name in
                      let tags =
                        List.sort tags
                          ~cmp:(
                            fun (tag1, _) (tag2, _) ->
                              compare (hash_variant tag1) (hash_variant tag2))
                      in
                      oc "/* %s : conversion table */\n" name;
                      let static = if not public then "static " else "" in
                      oc "%slookup_info ml_table_%s[] = {\n" static name;
                      oc "  { 0, %d },\n" (List.length tags);
                      List.iter tags
                        ~f:(
                           fun (tag, trans) ->
                             oc "  { MLTAG_%s, %s },\n" tag (ctag tag trans));
                      oc "};\n\n";
                      if not !first then oh "\n";
                      if public then
                        oh "extern lookup_info ml_table_%s[];\n" name;
                      oh "#define Val_%s(data) ml_lookup_from_c (ml_table_%s, data)\n"
                         name name;
                      oh "#define %s_val(key) ml_lookup_to_c (ml_table_%s, key)\n\n"
                         cname name
                  | _ -> raise (Stream.Error "")
                  end
              | _ -> raise (Stream.Error "")
              end
          | _ -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | _ -> raise End_of_file


let process ic ~hc ~cc =
  let chars = Stream.of_channel ic in
  let s = lexer chars in
  try while true do declaration s ~hc ~cc done with
    End_of_file -> ()
  | Stream.Error err ->
      failwith
        (Printf.sprintf "Parsing error \"%s\" at character %d on input stream"
           err (Stream.count chars))

let main () =
  let inputs = ref [] in
  let header = ref "" in
  let code = ref "" in
  Arg.parse
    ["-h", Arg.String (( := ) header), "file to output macros (file.h)";
     "-c", Arg.String (( := ) code),
     "file to output conversion tables (file.c)";
     "-static", Arg.Set static, "do not export conversion tables"]
    (fun s -> inputs := s :: !inputs) "usage: varcc [options] file.var";
  let inputs = List.rev !inputs in
  begin match inputs with
    [] ->
      if !header = "" then header := "a.h"; if !code = "" then code := "a.c"
  | ip :: _ ->
      let rad =
        if Filename.check_suffix ip ".var" then Filename.chop_extension ip
        else ip
      in
      if !header = "" then header := rad ^ ".h";
      if !code = "" then code := rad ^ ".c"
  end;
  let hc = open_out !header
  and cc = open_out !code in
  let chars = Stream.of_channel stdin in
  if inputs = [] then process stdin ~hc ~cc
  else
    List.iter inputs
      ~f:(
         fun file ->
           let ic = open_in file in
           try process ic ~hc ~cc; close_in ic with
             exn ->
               close_in ic; prerr_endline ("Error in " ^ file); raise exn);
  close_out hc;
  close_out cc

let _ = Printexc.print main ()
